
# Clustering of programming language features for Python
# - First, clustering is applied using Expectation Maximization
# - Second, a classification is identified (Common, IO, Multi, and Meta)
# - Third, statistically significance is studied
#
# Francisco Ortin
# University of Oviedo
# http://www.di.uniovi.es/~ortin


###### Clustering #########

library(mclust)

# Unidimensional clustering using EM
classifyUnidim <- function(data, classes=-1) {
  len <- length(data)
  if (classes==-1) 
    model <- Mclust(data) 
  else  {
    model <- Mclust(data, classes)
  }
  thematrix <- matrix(nrow = len, ncol = 2, byrow = TRUE)
  for (i in 1:len) {
    thematrix[i,1] <- i
    thematrix[i,2] <- data[i]
  }
  plot(model, thematrix, what=c('classification'))
  summary(model, parameters=TRUE)
  model
}


classifyCSV <- function(filename, column) {
  cat("\n\nClustering for ", filename, ":", sep="")
  data <- read.csv(file = filename, head=TRUE, sep=",")
  data <- na.omit(data)
  model <- classifyUnidim(data[,column])
  print(table(data$Classification, model$classification))
  model
}

# startup (python2)
model <- classifyCSV("cython-python2-micro-startup.csv", 2)
model <- classifyCSV("stackless-python2-micro-startup.csv", 2)
model <- classifyCSV("wpython-python2-micro-startup.csv", 2)
model <- classifyCSV("pypy-python2-micro-startup.csv", 2)
model <- classifyCSV("jython-python2-micro-startup.csv", 2)
model <- classifyCSV("iron-python2-micro-startup.csv", 2)

# steady (python2)
model <- classifyCSV("cython-python2-micro-steady.csv", 2)
model <- classifyCSV("stackless-python2-micro-steady.csv", 2)
model <- classifyCSV("wpython-python2-micro-steady.csv", 2)
model <- classifyCSV("pypy-python2-micro-steady.csv", 2)
model <- classifyCSV("jython-python2-micro-steady.csv", 2)
model <- classifyCSV("iron-python2-micro-steady.csv", 2)

# startup (python3)
model <- classifyCSV("cython-python3-micro-startup.csv", 2)
model <- classifyCSV("stackless-python3-micro-startup.csv", 2)
model <- classifyCSV("pypy-python3-micro-startup.csv", 2)

# steady (python3)
model <- classifyCSV("cython-python3-micro-steady.csv", 2)
model <- classifyCSV("stackless-python3-micro-steady.csv", 2)
model <- classifyCSV("pypy-python3-micro-steady.csv", 2)


###### Significant difference #########

# A function returning the CI of a distribution (vector)
computeCI <- function(dist, CL=0.95) {
  bootstrap <- c() # empty vector
  for (i in 1:1000) {
    sample10 <- sample(dist, 10, replace=T)
    mean <- mean(sample10)
    bootstrap <- c(bootstrap, mean)
  }
  alpha = (1-CL)/2
  lowCI <- quantile(bootstrap, alpha)
  hiCI <- quantile(bootstrap, 1-alpha)
  c(lowCI, hiCI)
}

# to compute whether two intervals overlap
doCIsOverlap <- function(ci1, ci2) max(ci1[1], ci2[1]) < min(ci1[2], ci2[2])

# computes the confidence level of two distributions that consider them as significantly different
getCL4SigDifferent <- function(dist1, dist2) {
  dist1 <- na.omit(dist1)
  dist2 <- na.omit(dist2)
  for(confidenceLevel in seq(0.99, 0.01, by=-0.01)) {
    dist1CI <- computeCI(dist1, confidenceLevel)
    dist2CI <- computeCI(dist2, confidenceLevel)
    if (!doCIsOverlap(dist1CI, dist2CI)) 
      break
  }
  confidenceLevel
}

# show significant difference of the four classifications, given a CSV file
showSignificances <- function(filename, minimumIL=0.5) {
  show <- function(name1, name2, CL) {
    if (CL<minimumIL) cat("  ", name1, " and ", name2, " are not significantly different.\n", sep="")
    else cat("  ", name1, " and ", name2, " are significantly different with probability of ", CL, ".\n", sep="")
  }
  cat("\n\nSignificance in ", filename, ":\n",  sep="")
  data <- read.csv(file = filename, head=TRUE, sep=",")
  common <- data[data$Classification=="Common",2]
  io <- data[data$Classification=="IO",2]
  multi <- data[data$Classification=="Multi",2]
  meta <- data[data$Classification=="Meta",2]
  show("IO", "Common", getCL4SigDifferent(common, io))
  show("Multi", "Common", getCL4SigDifferent(common, multi))
  show("Meta", "Common", getCL4SigDifferent(common, meta))
}

# Startup (python2)
showSignificances("cython-python2-micro-startup.csv")
showSignificances("stackless-python2-micro-startup.csv")
showSignificances("wpython-python2-micro-startup.csv")
showSignificances("pypy-python2-micro-startup.csv")
showSignificances("jython-python2-micro-startup.csv")
showSignificances("iron-python2-micro-startup.csv")


# Steady (python2)
showSignificances("cython-python2-micro-steady.csv")
showSignificances("stackless-python2-micro-steady.csv")
showSignificances("wpython-python2-micro-steady.csv")
showSignificances("pypy-python2-micro-steady.csv")
showSignificances("jython-python2-micro-steady.csv")
showSignificances("iron-python2-micro-steady.csv")

# Startup (python3)
showSignificances("cython-python3-micro-startup.csv")
showSignificances("stackless-python3-micro-startup.csv")
showSignificances("pypy-python3-micro-startup.csv")

# Steady (python3)
showSignificances("cython-python3-micro-steady.csv")
showSignificances("stackless-python3-micro-steady.csv")
showSignificances("pypy-python3-micro-steady.csv")

